package backupmanager;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

public class BackupManagerNasWin implements BackupManager {

	private Properties props = null;

	public BackupManagerNasWin(Properties props) throws IOException, InterruptedException, BackupManagerException {
		this.props = props;
		this.init();
	}

	private void init() throws IOException, InterruptedException, BackupManagerException {
		System.out.println("Initializing BackupManagerNasWin...");

		this.checkConfiguration();

		System.out.println("Mounting NAS shared folders...");
		for(String sharedFolder : this.props.getProperty("nas.sharedfolderslist").split("\\|")) {
			this.execCommand("/C", "net use " +  sharedFolder.split(">")[1] + " \\\\" + this.props.getProperty("nas.hostname") + "\\" + sharedFolder.split(">")[0]);

			String bkMngDirPath = sharedFolder.split(">")[1] + "/" + BackupManagerConsts.BKMNGDIRPATH;
			File bkMngDir = new File(bkMngDirPath);

			if(!bkMngDir.exists()) {
				this.execCommand("/C", "mkdir " + bkMngDirPath.replace("/","\\"));
			} else if(!bkMngDir.isDirectory()) {
				throw new BackupManagerException("NAS BKMNGDIR " + bkMngDirPath + " is not a folder.");
			}
		}
		System.out.println("NAS shared folders mounted.");

		System.out.println("BackupManagerNasWin initialized.");
	}

	private void checkConfiguration() throws BackupManagerException {
		System.out.println("Checking configuration...");
		
		List<String> checkList = new ArrayList<String>();
		
		for(String sharedFolder : this.props.getProperty("nas.sharedfolderslist").split("\\|")) {
			if(!checkList.contains(sharedFolder.split(">")[0])) {
				checkList.add(sharedFolder.split(">")[0]);
			} else {
				throw new BackupManagerException("CONFIGURATION: Duplicate NAS shared folder: " + sharedFolder.split(">")[0]);
			}
		}
		
		checkList.removeAll(checkList);
		
		for(String sharedFolder : this.props.getProperty("nas.sharedfolderslist").split("\\|")) {
			if(!checkList.contains(sharedFolder.split(">")[1])) {
				checkList.add(sharedFolder.split(">")[1]);
			} else {
				throw new BackupManagerException("CONFIGURATION: Duplicate NAS unit: " + sharedFolder.split(">")[1]);
			}
		}
		
		if(!this.props.getProperty("master.bkfolderslist").equals("")) {
			checkList.removeAll(checkList);
			
			for(String prop : this.props.getProperty("master.bkfolderslist").split("\\|")) {
				if(!checkList.contains(prop)) {
					checkList.add(prop);
				} else {
					throw new BackupManagerException("CONFIGURATION: Duplicate bkfolder: " + prop);
				}
			}
			
			Map<String,List<String>> checkMap = new HashMap<String,List<String>>();
			
			for(String prop : this.props.getProperty("master.bkfolderslist").split("\\|")) {
				String key = prop.split(">")[1].substring(2);
				String value = prop.split(">")[1].substring(0,2);
				if(!checkMap.containsKey(key)) {
					checkMap.put(key,new ArrayList<String>());
				}
				if(!checkMap.get(key).isEmpty() && !checkMap.get(key).contains(value)) {
					throw new BackupManagerException("CONFIGURATION: Duplicate master folder (from different drives): " + key);
				} else {
					checkMap.get(key).add(value);
				}
			}
		}
		
		System.out.println("Configuration checked.");
	}

	private void execCommand(String arg1, String arg2) throws IOException, InterruptedException, BackupManagerException {
		System.out.println("#####Executing command cmd.exe: " + arg1 + " " + arg2 + "...");
		ProcessBuilder pb = new ProcessBuilder("cmd.exe", arg1, arg2);
		pb.redirectErrorStream(true);
		Process p = pb.start();

		BufferedReader r = new BufferedReader(new InputStreamReader(p.getInputStream()));
		String line = null;
		while (true) {
			line = r.readLine();
	        if (line == null) { break; }
	        System.out.println("#####" + line);
		}
		r.close();

		int exitValue = p.waitFor();
		System.out.println("#####Command executed with exit value: " + exitValue + ".");
		if (exitValue != 0) {
			throw new BackupManagerException("execCommand failed.");
		}
	}

	public void synchronize(boolean forceUpdate, boolean forceRemove) throws BackupManagerException, IOException, InterruptedException {
		System.out.println("Synchronizing backup...");
		
		Map<String,List<String[]>> synchActionMap = this.checkSynch();
		
		System.out.println("Creating new folders...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_NEW_FOLDERS)) {
			this.execCommand("/C", "mkdir \"" + action[1].replace("/","\\") + "\"");
		}
		System.out.println("New folders created.");
		
		System.out.println("Copying new files...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_NEW_FILES)) {
			this.execCommand("/C", "copy /Y \"" + action[0].replace("/","\\") + "\" \"" + action[1].replace("/","\\") + "\"");
		}
		System.out.println("New files copied.");
		
		System.out.println("Overwriting updated files...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_UPDATED_FILES)) {
			this.execCommand("/C", "copy /Y \"" + action[0].replace("/","\\") + "\" \"" + action[1].replace("/","\\") + "\"");
		}
		System.out.println("Updated files overwrited.");
		
		System.out.println("Overwriting outdated files?...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_OUTDATED_FILES)) {
			if(forceUpdate) {
				this.execCommand("/C", "copy /Y \"" + action[0].replace("/","\\") + "\" \"" + action[1].replace("/","\\") + "\"");
			} else {
				System.out.println("WARNING: " + BackupManagerConsts.ACTION_OUTDATED_FILES + ":" + action[0] + "-->" + action[1] + ".");
				System.out.println("Use forceUpdate parameter to force update.");
			}
		}
		System.out.println("Outdated files overwrited?.");
		
		System.out.println("Removing deleted files?...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FILES)) {
			if(forceRemove) {
				this.execCommand("/C", "del \"" + action[1].replace("/","\\") + "\"");
			} else {
				System.out.println("WARNING: " + BackupManagerConsts.ACTION_DELETED_FILES + ":" + action[0] + "-->" + action[1] + ".");
				System.out.println("Use forceRemove parameter to force remove.");
			}
		}
		System.out.println("Deleted files removed?.");
		
		System.out.println("Removing deleted folders?...");
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FOLDERS)) {
			if(forceRemove) {
				this.execCommand("/C", "rmdir \"" + action[1].replace("/","\\") + "\"");
			} else {
				System.out.println("WARNING: " + BackupManagerConsts.ACTION_DELETED_FOLDERS + ":" + action[0] + "-->" + action[1] + ".");
				System.out.println("Use forceRemove parameter to force remove.");
			}
		}
		System.out.println("Deleted folders removed?.");
		
		System.out.println("Backup synchronized.");
	}

	public void printSynch() throws BackupManagerException, IOException {
		System.out.println("Printing synch...");
		
		Map<String,List<String[]>> synchActionMap = this.checkSynch();
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_NEW_FOLDERS)) {
			System.out.println(BackupManagerConsts.ACTION_NEW_FOLDERS + ":" + action[0].replace("/","\\") + "-->" + action[1].replace("/","\\") + ".");
		}
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_NEW_FILES)) {
			System.out.println(BackupManagerConsts.ACTION_NEW_FILES + ":" + action[0].replace("/","\\") + "-->" + action[1].replace("/","\\") + ".");
		}
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_UPDATED_FILES)) {
			System.out.println(BackupManagerConsts.ACTION_UPDATED_FILES + ":" + action[0].replace("/","\\") + "-->" + action[1].replace("/","\\") + ".");
		}
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_OUTDATED_FILES)) {
			System.out.println(BackupManagerConsts.ACTION_OUTDATED_FILES + ":" + action[0].replace("/","\\") + "-->" + action[1].replace("/","\\") + ".");
		}
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FILES)) {
			System.out.println(BackupManagerConsts.ACTION_DELETED_FILES + ":" + action[0] + "-->" + action[1].replace("/","\\") + ".");
		}
		
		for(String[] action : synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FOLDERS)) {
			System.out.println(BackupManagerConsts.ACTION_DELETED_FOLDERS + ":" + action[0] + "-->" + action[1].replace("/","\\") + ".");
		}
		
		System.out.println("Synch printed.");
	}

	private Map<String,List<String[]>> checkSynch() throws BackupManagerException, IOException {
		System.out.println("Checking synch...");
		
		Map<String,List<String[]>> synchActionMap = new HashMap<String,List<String[]>>();
		synchActionMap.put(BackupManagerConsts.ACTION_NEW_FOLDERS,new ArrayList<String[]>());
		synchActionMap.put(BackupManagerConsts.ACTION_NEW_FILES,new ArrayList<String[]>());
		synchActionMap.put(BackupManagerConsts.ACTION_UPDATED_FILES,new ArrayList<String[]>());
		synchActionMap.put(BackupManagerConsts.ACTION_OUTDATED_FILES,new ArrayList<String[]>());
		synchActionMap.put(BackupManagerConsts.ACTION_DELETED_FOLDERS,new ArrayList<String[]>());
		synchActionMap.put(BackupManagerConsts.ACTION_DELETED_FILES,new ArrayList<String[]>());
		
		if(!this.props.getProperty("master.bkfolderslist").equals("")) {
			for(String prop : this.props.getProperty("master.bkfolderslist").split("\\|")) {
				this.checkMasterFolder(prop.split(">")[1], prop.split(">")[0], synchActionMap);
			}
		}
		
		for(String prop : this.props.getProperty("nas.sharedfolderslist").split("\\|")) {
			this.checkNasUnit(prop.split(">")[1], synchActionMap);
		}
		
		Collections.reverse(synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FOLDERS));
		
		System.out.println("Synch checked.");
		
		return synchActionMap;
	}

	private Map<String,List<String[]>> checkMasterFolder(String masterFolderPath, String nasUnit, Map<String,List<String[]>> synchActionMap) throws BackupManagerException, IOException {
		
		String nasFolderPath = nasUnit + "/" + BackupManagerConsts.BKMNGDIRPATH + masterFolderPath.substring(2);
		
		File masterFolder = new File(masterFolderPath);
		File nasFolder = new File(nasFolderPath);
		
		if(!masterFolder.exists() || !masterFolder.isDirectory()) {
			throw new BackupManagerException("Master folder " + masterFolderPath + " does not exists or is not a folder.");
		}
		
		if((nasFolder.exists() && nasFolder.isDirectory()) || !nasFolder.exists()) {
			
			if(!nasFolder.exists()) {
				/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_NEW_FOLDERS)) {
					synchActionMap.put(BackupManagerConsts.ACTION_NEW_FOLDERS,new ArrayList<String[]>());
				}*/
				synchActionMap.get(BackupManagerConsts.ACTION_NEW_FOLDERS).add(new String[]{masterFolderPath,nasFolderPath});
			}
			
			for(File masterFile : masterFolder.listFiles()) {
				if(masterFile.isHidden()) {continue;}
				if(masterFile.isDirectory()) {
					this.checkMasterFolder(masterFile.getCanonicalPath(), nasUnit, synchActionMap);
				} else {
					String nasFilePath = nasUnit + "/" + BackupManagerConsts.BKMNGDIRPATH + masterFile.getCanonicalPath().substring(2);
					File nasFile = new File(nasFilePath);
					
					if((nasFile.exists() && !nasFile.isDirectory()) || !nasFile.exists()) {
						
						if(!nasFile.exists()) {
							/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_NEW_FILES)) {
								synchActionMap.put(BackupManagerConsts.ACTION_NEW_FILES,new ArrayList<String[]>());
							}*/
							synchActionMap.get(BackupManagerConsts.ACTION_NEW_FILES).add(new String[]{masterFile.getCanonicalPath(),nasFilePath});
						} else if (masterFile.lastModified() - nasFile.lastModified() > 0) {
							/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_UPDATED_FILES)) {
								synchActionMap.put(BackupManagerConsts.ACTION_UPDATED_FILES,new ArrayList<String[]>());
							}*/
							synchActionMap.get(BackupManagerConsts.ACTION_UPDATED_FILES).add(new String[]{masterFile.getCanonicalPath(),nasFilePath});
						} else if (masterFile.length() != nasFile.length()) {
							/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_OUTDATED_FILES)) {
								synchActionMap.put(BackupManagerConsts.ACTION_OUTDATED_FILES,new ArrayList<String[]>());
							}*/
							synchActionMap.get(BackupManagerConsts.ACTION_OUTDATED_FILES).add(new String[]{masterFile.getCanonicalPath(),nasFilePath});
						}
						
					} else {
						throw new BackupManagerException("NAS file " + nasFilePath + " is not a file.");
					}
				}
			}
			
		} else {
			throw new BackupManagerException("NAS folder " + nasFolderPath + " is not a folder.");
		}
		
		return synchActionMap;
		
	}

	private Map<String,List<String[]>> checkNasUnit(String nasUnit, Map<String,List<String[]>> synchActionMap) throws BackupManagerException, IOException {
		
		String bkMngDirPath = nasUnit + "/" + BackupManagerConsts.BKMNGDIRPATH;
		File bkMngDir = new File(bkMngDirPath);
		
		if(bkMngDir.exists() && bkMngDir.isDirectory()) {
			
			for(File nasFile : bkMngDir.listFiles()) {
				
				boolean folderConfigured = false;
				
				if(nasFile.isDirectory()) {
					if(!this.props.getProperty("master.bkfolderslist").equals("")) {
						for(String prop : this.props.getProperty("master.bkfolderslist").split("\\|")) {
							
							String masterFolder = prop.split(">")[1].substring(2);
							
							if(prop.split(">")[1].indexOf('/',3) != -1) {
								masterFolder = prop.split(">")[1].substring(2,prop.split(">")[1].indexOf('/',3));
							}
							
							if(nasUnit.equals(prop.split(">")[0]) && masterFolder.equals(nasFile.getCanonicalPath().substring(bkMngDirPath.length()).replace("\\","/"))) {
								folderConfigured = true;
								this.checkNasFolder(nasFile.getCanonicalPath(), prop.split(">")[1].substring(0, 2) + masterFolder, nasUnit, synchActionMap);
								break;
							}
						}
					}
					
					if(!folderConfigured) {
						/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_DELETED_FOLDERS)) {
							synchActionMap.put(BackupManagerConsts.ACTION_DELETED_FOLDERS,new ArrayList<String[]>());
						}*/
						//synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FOLDERS).add(new String[]{null,nasFile.getCanonicalPath()});
						this.checkNasFolder(nasFile.getCanonicalPath(), null, nasUnit, synchActionMap);
					}
				} else {
					throw new BackupManagerException("NAS BKMNGDIR " + bkMngDirPath + " contains a file: " + nasFile.getCanonicalPath() + ".");
				}
			}
			
		} else {
			throw new BackupManagerException("NAS BKMNGDIR " + bkMngDirPath + " does not exists or is not a folder.");
		}
		
		return synchActionMap;
		
	}

	private Map<String,List<String[]>> checkNasFolder(String nasFolderPath, String masterFolderPath, String nasUnit, Map<String,List<String[]>> synchActionMap) throws BackupManagerException, IOException {
		
		File nasFolder = new File(nasFolderPath);
		File masterFolder = (masterFolderPath == null ? null : new File(masterFolderPath));
		
		if(masterFolder == null || (masterFolder.exists() && masterFolder.isDirectory()) || !masterFolder.exists()) {
			
			if(masterFolder == null || !masterFolder.exists()) {
				/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_DELETED_FOLDERS)) {
					synchActionMap.put(BackupManagerConsts.ACTION_DELETED_FOLDERS,new ArrayList<String[]>());
				}*/
				synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FOLDERS).add(new String[]{null,nasFolderPath});
			}
			
			for(File nasFile : nasFolder.listFiles()) {
				if(nasFile.isDirectory()) {
					this.checkNasFolder(nasFile.getCanonicalPath(), (masterFolder == null ? null : masterFolder + "/" + nasFile.getName()), nasUnit, synchActionMap);
				} else {
					String masterFilePath = (masterFolderPath == null ? null : masterFolderPath + "/" + nasFile.getName());
					File masterFile = (masterFilePath == null ? null : new File(masterFilePath));
					
					if(masterFile == null || !masterFile.exists()) {
						/*if(!synchActionMap.containsKey(BackupManagerConsts.ACTION_DELETED_FILES)) {
							synchActionMap.put(BackupManagerConsts.ACTION_DELETED_FILES,new ArrayList<String[]>());
						}*/
						synchActionMap.get(BackupManagerConsts.ACTION_DELETED_FILES).add(new String[]{null,nasFile.getCanonicalPath()});
					} else if(masterFile.isDirectory()) {
						throw new BackupManagerException("Master file " + masterFilePath + " is not a file.");
					}
				}
			}
			
		} else {
			throw new BackupManagerException("Master folder " + masterFolderPath + " is not a folder.");
		}
		
		return synchActionMap;
		
	}

	public void shutDown() throws IOException, InterruptedException, BackupManagerException {
		System.out.println("Ending BackupManagerNasWin...");

		System.out.println("Unmounting NAS shared folders...");
		for(String sharedFolder : this.props.getProperty("nas.sharedfolderslist").split("\\|")) {
			this.execCommand("/C", "net use " +  sharedFolder.split(">")[1] + " /delete");
		}
		System.out.println("NAS shared folders unmounted.");

		System.out.println("BackupManagerNasWin Ended.");
	}

}
